#!/usr/bin/env python

"""

   web-malware-scanner.py,

        A Yara wrapper to efficiently detect malware code in
        web files. Supports incremental mode, whitelisting and
        extension filtering.

        See `mwscan.py --help` for usage.

   Copyright (C) 2017 Willem de Groot <gwillem@gmail.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

"""

import os
import sys
import argparse
import logging
import hashlib
import time
from os import path
from ruleset import providers

try:
    import psutil
except ImportError:
    psutil = None

DEFAULT_RULES_FILE = path.join(path.dirname(__file__), 'data', 'all-confirmed.yar')
LAST_RUN_FILE = path.expanduser('~/.mwscan_last_run')
CODE_EXT = ('php', 'phtml', 'js', 'jsx', 'html', 'php3', 'php4', 'php5', 'php7', 'sh')

logging.basicConfig(format='%(asctime)s %(message)s', datefmt='%c', level=logging.DEBUG)
logging.getLogger('requests').setLevel(logging.WARNING)


def parse_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Scan webfolders for hidden malware.",
    )
    parser.add_argument('path', help='File or dir to scan.')
    parser.add_argument('-n', '--newonly', help='Only consider files that were modified/created since last succesful run.'.format(LAST_RUN_FILE), action='store_true')
    parser.add_argument('-d', '--deep', action='store_true', help='Scan all files, not just code.')
    parser.add_argument('-q', '--quiet', action='store_true', help='Supress output except for matches.')
    parser.add_argument('-r', '--rules', help='Yara rules file.', default=DEFAULT_RULES_FILE)
    parser.add_argument('-s', '--ruleset', choices=sorted(providers.keys()), help='Download and use from upstream')

    args = parser.parse_args()

    if args.quiet:
        logging.root.setLevel(logging.WARNING)

    if (args.ruleset and args.ruleset != 'file'):
        args.rules = None

    if args.rules:
        args.ruleset = 'file'

    if args.newonly and not os.path.isdir(args.path):
        logging.error('I can only use --newonly on a directory!')
        sys.exit(1)

    if not os.path.exists(args.path):
        logging.error('{0} does not exist!'.format(args.path))
        sys.exit(1)

    if args.deep:
        args.req_ext = []
    else:
        args.req_ext = CODE_EXT

    if args.rules and not os.path.isfile(args.rules):
        logging.error("Rules file {0} does not exist. Use --rules <path>".format(args.rules))
        sys.exit(1)

    if args.newonly:
        args.newer_than = find_last_run_timestamp(args.path)
    else:
        args.newer_than = None

    return args


def path_to_timestamp_file(path):

    id = str(os.geteuid()) + '\0' + path
    hash = hashlib.md5(id.encode()).hexdigest()
    return LAST_RUN_FILE + '.' + hash


def find_last_run_timestamp(path):
    tsfile = path_to_timestamp_file(path)

    if os.path.isdir(path) and os.path.isfile(tsfile):
        ts = os.path.getmtime(tsfile)
        logging.debug("Not scanning files older than {0} (from {1})".format(time.ctime(ts), tsfile))
    else:
        ts = None
        logging.debug("No timestamp reference file found for {0}, I will scan everything.".format(path))

    return ts


def write_last_run_timestamp(path):

    tsfile = path_to_timestamp_file(path)
    logging.debug("Storing last run timestamp for {0} in {1}".format(path, tsfile))

    if os.path.exists(tsfile):
        os.unlink(tsfile)

    return open(tsfile, 'a').close()


def find_targets(root_path, newer_than=None, req_ext=None):
    """
     Produce an iterator for all the relevant files recursively found under root_path.
     If root_path is a file, return an iterator with just that.
     Param ignore_older_than is an epoch timestamp from os.path.getmtime()
    """

    if os.path.isfile(root_path):
        yield root_path
    elif not os.path.isdir(root_path):
        raise RuntimeError("%s is not a file or directory" % root_path)

    for root, dirs, files in os.walk(root_path):
        for myfile in files:
            path = os.path.join(root, myfile)

            if req_ext and myfile.split('.')[-1] not in req_ext:
                    continue

            if newer_than:
                # if mtime OR ctime is newer, do scan
                if os.path.getmtime(path) < newer_than and os.path.getctime(path) < newer_than:
                    continue

            if os.path.islink(path):
                continue

            yield path


def scan_files(files, rules, whitelist):
    num_files = 0
    num_malware = 0
    num_whitelisted = 0

    for path in files:
        num_files += 1

        with open(path, 'rb') as fh:
            data = fh.read()

        sha1sum = hashlib.sha1(data).hexdigest()
        if sha1sum in whitelist:
            num_whitelisted += 1
            continue

        matches = rules.match(data=data)
        if matches:
            num_malware += 1

        for m in matches:
            print("{0!s} {1}".format(m, path))

    return num_files, num_malware, num_whitelisted


def main():

    args = parse_args()

    # don't swamp the machine
    if psutil:
        mylife = psutil.Process()
        mylife.ionice(psutil.IOPRIO_CLASS_IDLE)

    provider = providers[args.ruleset]
    rules, whitelist = provider(args=args).get()

    try:
        files = find_targets(args.path,
                             newer_than=args.newer_than,
                             req_ext=args.req_ext)

        total, malware, whitelisted = scan_files(files, rules, whitelist)

        if args.newonly:
            write_last_run_timestamp(args.path)

        logging.info("Finished scanning {0} files: {1} malware and {2} whitelisted.".format(total, malware, whitelisted))

    except KeyboardInterrupt:
        return 1


if __name__ == '__main__':
    sys.exit(main())



import requests
import re
import json
import logging
import yara


def strip_last_url_path(url):
    parent, _, _ = url.rpartition('/')
    return parent


def last_url_path(url):
    return url.rpartition('/')[2]


class RulesProvider:
    rules_url = None
    whitelist_url = None

    def __init__(self, **kwargs):
        logging.info("Using {0} rules.".format(self.__class__.__name__))
        self._args = kwargs.get('args')

    def find_whitelist_in_rawrules(self, rawrules):
        # Find whitelist hashes from comments, because yara whitelist
        # hashing is too slow. See https://github.com/VirusTotal/yara/issues/592

        m = re.search('/\*[^*]*WHITELIST = (\{.*?\})\s*\*/', rawrules, flags=re.DOTALL)
        return set(json.loads(m.group(1)) if m else [])

    def get_rules(self):
        return self._recursive_fetch(self.rules_url)

    def get_whitelist(self):
        if not self.whitelist_url:
            return set()

        data = self._httpget(self.whitelist_url)
        hashes = re.findall('[a-f0-9]{40}', data)  # assume sha1 hex hash
        return set(hashes)

    def transform_rules(self, rawrules):
        """For specific rules providers, to mangle into mwscan compatible form"""
        whitelist = set()
        return rawrules, whitelist


    def _httpget(self, url):

        filename = last_url_path(url)
        logging.debug("Fetching {0}".format(filename))

        return requests.get(url).text


    def get(self):
        """Returns rules, whitelist"""

        rawrules = self.get_rules()

        # provider specific transformation
        rawrules, whitelist = self.transform_rules(rawrules)

        # if alternative whitelist method is required
        whitelist.update(self.get_whitelist())
        whitelist.update(self.find_whitelist_in_rawrules(rawrules))

        rules = yara.compile(source=rawrules)

        logging.debug("Loaded {0} yara rules and {1} whitelist entries".format(
            len(list(iter(rules))),
            len(whitelist),
        ))

        return rules, whitelist

    def _recursive_fetch(self, url):

        def include(match):
            relpath = match.group(1)
            # return match.group(1)
            newurl = strip_last_url_path(url) + '/' + relpath
            return "/* included from {0} */\n".format(newurl) + self._recursive_fetch(newurl)

        data = self._httpget(url)
        data = re.sub('include "([^"]+?)"\s+', include, data)
        # data = re.sub('import "hash"\s*', '', data)
        return data


class Files(RulesProvider):

    # initialize with Files(args)
    def get_rules(self):
        path = self._args.rules
        logging.info("Loading {0}".format(self._args.rules))
        with open(path) as fh:
            return fh.read()


class NBS(RulesProvider):

    rules_url = 'https://raw.githubusercontent.com/nbs-system/php-malware-finder/master/php-malware-finder/php.yar'

    def transform_rules(self, rawrules):
        whitelist = set()
        rules = list()

        tokens = re.findall('\n(?:global )?(?:private )?rule .+?\n\{\n.+?\n\}', rawrules, flags=re.DOTALL)

        for token in tokens:
            hashes = re.findall('== "([a-f0-9]{40})"', token)

            if 'rule IsWhitelisted' in token:
                continue

            if hashes or 'hash.sha1' in token:
                whitelist.update(hashes)
            else:
                token = token.strip()
                token = re.sub(' and not IsWhitelisted', '', token)

                rules.append(token.strip())

        return '\n'.join(rules), whitelist


class Magemojo(RulesProvider):
    rules_url = 'https://raw.githubusercontent.com/magesec/magesecurityscanner/master/yararules.yar'
    whitelist_url = 'https://raw.githubusercontent.com/magesec/magesecurityscanner/master/magesecurityscan/sha1whitelist.json'


class Magesec(RulesProvider):
    rules_url = 'https://magesec.org/download/yara-standard.yar'
    whitelist_url = 'https://magesec.org/download/whitelist.json'


class Byte(RulesProvider):
    rules_url = 'https://raw.githubusercontent.com/gwillem/magento-malware-scanner/master/build/all-confirmed.yar'


providers = {
    'nbs': NBS,
    'byte': Byte,
    'magemojo': Magemojo,
    'magesec': Magesec,
    'file': Files,
}



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup
from os import path

here = path.abspath(path.dirname(__file__))

VERSION="20170117.194808"

setup(
    name='mwscan',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=VERSION,

    description='Find malware in web documents.',
    long_description='A Yara wrapper to identify malware, with support for incremental scans, extension filtering and efficient hash whitelisting.',

    # The project's main homepage.
    url='https://github.com/gwillem/magento-malware-scanner',
    download_url='https://github.com/gwillem/magento-malware-scanner/tarball/{0}'.format(VERSION),

    # Author details
    author='Willem de Groot',
    author_email='gwillem@gmail.com',

    # Choose your license
    license='GPLv3',

    # # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    # classifiers=[
    #     # How mature is this project? Common values are
    #     #   3 - Alpha
    #     #   4 - Beta
    #     #   5 - Production/Stable
    #     'Development Status :: 4 - Beta',

    #     # Indicate who your project is intended for
    #     'Intended Audience :: Developers',
    #     'Topic :: Software Development :: Build Tools',

    #     # Pick your license as you wish (should match "license" above)
    #     'License :: OSI Approved :: MIT License',

    #     # Specify the Python versions you support here. In particular, ensure
    #     # that you indicate whether you support Python 2, Python 3 or both.
    #     'Programming Language :: Python :: 2',
    #     'Programming Language :: Python :: 2.6',
    #     'Programming Language :: Python :: 2.7',
    #     'Programming Language :: Python :: 3',
    #     'Programming Language :: Python :: 3.3',
    #     'Programming Language :: Python :: 3.4',
    #     'Programming Language :: Python :: 3.5',
    # ],

    # What does your project relate to?
    keywords=['yara', 'malware', 'magento'],

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    # package[s=find_packages(exclude=['contrib', 'docs', 'tests']),
    packages=['mwscan'],

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['psutil', 'yara-python', 'requests'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    # extras_require={
    #     'dev': ['check-manifest'],
    #     'test': ['coverage'],
    # },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    # package_data={
    #     'sample': ['package_data.dat'],
    # },

    package_data={'mwscan': ['data/all-confirmed.yar']},

    # data_files = [
    #     ('data', ['build/all-confirmed.yar']),
    # ],

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'mwscan=mwscan.mwscan:main',
        ],
    },
)
